from datetime import datetime
import json
import os
import time
import random
import re
import shutil
import string
from subprocess import run, Popen, PIPE, STDOUT

import requests


class MultisigWallet():
    address = None
    cli_path = None
    errors = []
    finalized = False
    is_blocked = False
    key1 = None
    key2 = None
    key3 = None
    label = None
    name = None
    password = None
    port = None
    process = None
    rpc_path = None
    settings = {}

    def __init__(self, name, settings):
        self.name = str(name)
        self.port = random.randint(30000, 50000)
        self.settings = settings
        self.default_path = os.path.abspath(os.curdir)
        self.wallet_path = os.path.abspath(os.path.join('wallets', self.name))
        self.cli_path = os.path.abspath('bin/monero-wallet-cli')
        self.rpc_path = os.path.abspath('bin/monero-wallet-rpc')
        if not os.path.isdir(self.wallet_path):
            raise NameError(f'{self.wallet_path} is not found')
        os.chdir(self.wallet_path)
        self.password = open('wallet.password.txt').read().strip()
        if os.path.exists('wallet.label.txt'):
            self.label = open('wallet.label.txt').read().strip()
        if os.path.exists('ms.round1.txt'):
            self.key1 = open('ms.round1.txt').read().strip()
        if os.path.exists('ms.round2.txt'):
            self.key2 = open('ms.round2.txt').read().strip()
        if os.path.exists('ms.round3.txt'):
            self.key3 = open('ms.round3.txt').read().strip()
        if os.path.exists('ms.finalized.txt'):
            self.finalized = open('ms.finalized.txt').read().strip() == 'done'

    def __del__(self):
        self.rpc_stop()

    def get_cli_command(self, command=None, ignore_daemon=False):
        result = [
            self.cli_path,
            '--wallet-file=wallet',
            f'--password={self.password}',
            '--daemon-ssl-allow-any-cert',
        ]
        for key, value in self.settings.items():
            if ignore_daemon and key.startswith('daemon'):
                continue
            if value:
                if isinstance(value, str):
                    result.append(f'--{key}={value}')
                elif isinstance(value, bool) and value:
                    result.append(f'--{key}')
        if command:
            result.append(f'--command={command}')
        return result

    @classmethod
    def cli_command(self, command, input_text=os.linesep):
        process = Popen(command, stdout=PIPE, stdin=PIPE, stderr=STDOUT)
        resp = process.communicate(input=input_text.encode())
        process.terminate()
        time.sleep(5)
        process.kill()
        return resp[0].decode()

    def run_console(self):
        print(f'PASSWORD: {self.password}')
        self.rpc_stop()
        run(self.get_cli_command())

    def rpc_start(self, wait=False):
        print('Starting RPC, please wait...')
        os.chdir(self.wallet_path)
        command = [
            os.path.join(self.default_path, 'bin', 'monero-wallet-rpc'),
            '--wallet-file=wallet',
            f'--password={self.password}',
            f'--rpc-bind-port={self.port}',
            f'--rpc-login={self.name}:{self.password}',
            '--daemon-ssl-allow-any-cert',
        ]
        for key, value in self.settings.items():
            if value:
                if isinstance(value, str):
                    command.append(f'--{key}={value}')
                elif isinstance(value, bool) and value:
                    command.append(f'--{key}')
        self.process = Popen(command)#, stdout=PIPE, stdin=PIPE, stderr=STDOUT)
        self.auth = requests.auth.HTTPDigestAuth(self.name, self.password)
        self.session = requests.Session()
        self.url = f'http://127.0.0.1:{self.port}/json_rpc'
        if wait:
            i = 0
            while i < self.settings.get('timeout', 30):
                i += 1
                if self.test_command():
                    print('Connected...')
                    break
                else:
                    time.sleep(1)

    def rpc_stop(self):
        try:
            if self.process:
                print('Stopping RPC, please wait...')
                self.process.terminate()
                time.sleep(2)
                self.process.kill()
                self.process = None
        except Exception as e:
            print(f'Error: {e}')

    def test_command(self):
        try:
            resp = self.session.post(
                self.url,
                headers={'Content-Type': 'application/json'},
                data=json.dumps({'jsonrpc': '2.0', 'id': 0, 'method': 'get_version'}),
                auth=self.auth,
                timeout=self.settings.get('timeout', 30)
            )
            if resp.status_code == 200:
                return True
        except requests.exceptions.ConnectionError:
            ...
        return False

    def rpc_command(self, method, wait=True, **params):
        if not self.process:
            self.rpc_start(wait=wait)
        self.errors = []
        try:
            resp = self.session.post(
                self.url,
                headers={'Content-Type': 'application/json'},
                data=json.dumps({'jsonrpc': '2.0', 'id': 0, 'method': method, 'params': params or {}}),
                auth=self.auth,
                timeout=self.settings.get('timeout', 30)
            )
            if resp.status_code == 200:
                self.is_blocked = False
                try:
                    return resp.json().get('result')
                except Exception as e:
                    print('ERROR:', e)
                    ...
            self.errors.append(resp.content.decode())
        except (requests.exceptions.ReadTimeout, requests.exceptions.ConnectionError) as e:
            print('Blocking...')
            print(e)
            self.is_blocked = True
            self.errors.append('Wallet is busy')
        except Exception as e:
            print('ERROR:', e)
            self.errors.append(str(e))

    @classmethod
    def generate_wallet(cls, settings):
        cli_path = os.path.abspath('bin/monero-wallet-cli')
        if not os.path.isdir('wallets'):
            os.mkdir('wallets')
        max_num = 0
        for name in os.listdir('wallets'):
            dn = os.path.join('wallets', name)
            if os.path.isdir(dn):
                try:
                    if int(name) > max_num:
                        max_num = int(name)
                except Exception:
                    ...
        name = str(max_num + 1)
        dn = os.path.abspath(os.path.join('wallets', name))
        os.mkdir(dn)
        os.chdir(dn)
        password = ''.join([
            random.choice(string.ascii_letters)
            for x in range(16)
        ])
        command = [
            cli_path,
            '--generate-new-wallet=wallet',
            '--mnemonic-language=English',
            f'--password={password}',
            '--daemon-ssl-allow-any-cert',
        ]
        for key, value in settings.items():
            if value:
                if isinstance(value, str):
                    command.append(f'--{key}={value}')
                elif isinstance(value, bool) and value:
                    command.append(f'--{key}')
        resp = cls.cli_command(command, f'N{os.linesep}')
        try:
            viewkey = re.search(r'.*View key:\ (.+)', resp).group(1)
            seed = re.search(f'.*NOTE: the following 25 words can be used to recover access to your wallet. Write them down and store them somewhere safe and secure. Please do not store them in your email or on file storage services outside of your immediate control.{os.linesep}{os.linesep}(.+{os.linesep}.+{os.linesep}.+{os.linesep}).*', resp).group(1)  # noqa: E501
            open('wallet.viewkey.txt', 'w').write(viewkey)
            open('wallet.seed.txt', 'w').write(seed)
            open('wallet.password.txt', 'w').write(password)
        except Exception:
            shutil.rmtree(dn)
            print('ERROR:', resp)
        finally:
            os.chdir(os.path.join(os.pardir, os.pardir))
        return cls(name, settings=settings)

    def turn_ms_on(self):
        self.cli_command(
            self.get_cli_command(),
            f'set enable-multisig-experimental 1\n{self.password}',
        )

    def get_address(self):
        if not self.address:
            fn = os.path.join(self.wallet_path, 'wallet.address.txt')
            if os.path.exists(fn):
                self.address = open(fn).read().strip()
            else:
                data = self.rpc_command('get_address', account_index=0)
                if data:
                    self.address = data.get('address')
                    open(fn, 'w').write(self.address)
        return self.address

    def get_version(self):
        data = self.rpc_command('get_version')
        if data:
            return data['version']

    def balance(self):
        data = self.rpc_command('get_balance')
        if data:
            total = data["balance"] / 10**12
            unlocked = data["unlocked_balance"] / 10**12
            text = f'Balance: {total:.8f}'
            if unlocked < total:
                text = f'{text} (locked: {total-unlocked:.8f})'
            print(text)
            return {
                'total': total,
                'unlocked': unlocked,
            }
            return data

    def rescan(self):
        self.rpc_command('rescan_blockchain')
#         self.is_blocked = True

    def show_transfers(self):
        data = self.rpc_command('get_transfers', **{'in': True, 'out': True})
        if data:
            for item in data.get('in', []):
                lines = [
                    'IN',
                    datetime.fromtimestamp(item['timestamp']).strftime('%Y-%m-%d %H:%M:%S'),
                    f'txid: {item["txid"]}',
                    f'amount: {item["amount"] / 10**12:.8f}',
                ]
                if item["locked"]:
                    lines.append(f'LOCKED ({item["confirmations"]}/10)')
                print('\t'.join(lines))
            for item in data.get('out', []):
                lines = [
                    'OUT',
                    datetime.fromtimestamp(item['timestamp']).strftime('%Y-%m-%d %H:%M:%S'),
                    f'txid: {item["txid"]}',
                    f'amount: {item["amount"] / 10**12:.8f}',
                ]
                if item["locked"]:
                    lines.append(f'LOCKED ({item["confirmations"]}/10)')
                print('\t'.join(lines))
            return data

    def set_label(self, label):
        self.rpc_command('label_account', account_index=0, label=label)
        open(os.path.join(self.wallet_path, 'wallet.label.txt'), 'w').write(label)
        self.label = label

    def prepare_multisig(self):
        data = self.rpc_command('prepare_multisig')
        if data:
            self.key1 = data['multisig_info']
            open(os.path.join(self.wallet_path, 'ms.round1.txt'), 'w').write(self.key1)
            return self.key1

    def make_multisig(self, key1, key2):
        data = self.rpc_command(
            'make_multisig',
            multisig_info=[key1, key2],
            threshold=2,
            password=self.password,
        )
        if data:
            self.key2 = data['multisig_info']
            open(os.path.join(self.wallet_path, 'ms.round2.txt'), 'w').write(self.key2)
            return self.key2

    def exchange_multisig_keys(self, key1, key2):
        data = self.rpc_command(
            'exchange_multisig_keys',
            multisig_info=[key1, key2],
            password=self.password,
        )
        if data:
            self.key3 = data['multisig_info']
            open(os.path.join(self.wallet_path, 'ms.round3.txt'), 'w').write(self.key3)
            return self.key3

    def finalize_wallet(self, key1, key2):
        data = self.rpc_command(
            'exchange_multisig_keys',
            multisig_info=[key1, key2],
            password=self.password,
        )
        if data:
            self.address = data['address']
            open(os.path.join(self.wallet_path, 'ms.address.txt'), 'w').write(self.address)
            open(os.path.join(self.wallet_path, 'ms.finalized.txt'), 'w').write('done')
            self.finalized = True
            return self.address

    def export_multisig_info(self):
        data = self.rpc_command('export_multisig_info')
        if data:
            return data['info']

    def import_multisig_info(self, info):
        data = self.rpc_command('import_multisig_info', info=[info])
        if data:
            return data['n_outputs']

    def transfer(self, addr, amount):
        try:
            data = self.rpc_command(
                'transfer',
                destinations=[{
                    'amount': int(amount * 10**12),
                    'address': addr,
                }],
                get_tx_key=True,
                get_tx_hex=True,
                get_tx_metadata=True,
            )
            if data:
                return data['multisig_txset']
        except NameError:
            print('Balance is too small. Fee is not including to amount.')

    def show_multisig(self, tx):
        data = self.rpc_command('describe_transfer', multisig_txset=tx)
        if data:
            print('Destinations:')
            for item in data['desc'][0]['recipients']:
                print(f'{item["address"]}\t{item["amount"] / 10**12}')
            print(f'Fee: {data["desc"][0]["fee"]/10**12:.8f}')
            return data

    def sign_multisig(self, tx):
        data = self.rpc_command('sign_multisig', tx_data_hex=tx)
        if data:
            return data['tx_data_hex']

    def submit_multisig(self, tx):
        data = self.rpc_command('submit_multisig', tx_data_hex=tx)
        if data:
            return data['tx_hash_list']
