import os


class Template():
    def __init__(self, name):
        self.text = open(os.path.join('web', 'templates', name)).read()

    def start(self):
        return open(os.path.join('web', 'templates', 'start.html')).read()

    def end(self):
        return open(os.path.join('web', 'templates', 'end.html')).read()

    def render(self, context={}):
        text = self.start() + self.text + self.end()
        for key, value in context.items():
            text = text.replace('{{' + key + '}}', str(value))
        return text


class BaseView():
    template_name = None

    def get(self, request, *args, **kargs):
        return self.render()

    def post(self, request, *args, **kargs):
        return self.render()

    def render(self, context={}):
        return Template(self.template_name).render(context)

    def redirect(self, request, url):
        return request.redirect(url)

    @classmethod
    def get_form_data(cls, data):
        return data


class StaticView(BaseView):
    def get(self, request, name):
        try:
            return open(os.path.join('web', 'static', name)).read()
        except Exception as e:
            print(e)
            return 404, 'Not found'


class HomeView(BaseView):
    def get(self, request):
        return self.redirect(request, '/wallets/')


class WalletListView(BaseView):
    template_name = 'wallet-list.html'


class WalletDetailView(BaseView):
    template_name = 'wallet-detail.html'

    def get(self, request, wallet_id):
        return self.render({'wallet_id': wallet_id})
